/* This file can be redistributed and used under GNU Library General Public
 * License. (c) 2000 Micha Moskal <malekith@topnet.pl>. 
 * $Id: block.h,v 1.4 2000/09/09 17:59:14 malekith Exp $ */

#ifndef _V_BLOCK_H
#define _V_BLOCK_H

#include <v/utils.h>

XCPT(XBlock, XRecover);

#define BLOCK_MIN 4

template <class T>
class Block {
	T *data;
	int s_logic, s_real;
	void qsort(int, int);
public:
	T& operator [] (int n) 
	{ 
		if (n < 0 || n >= s_logic) 
			throw XBlock("operator []: index out of range");
		return data[n];
	}
	int push(const T &r);
	T pop();
	void free() { s_logic = s_real = 0; delete [] data; data = null; }
	int size() { return s_logic; }
	
	Block() : data(null), s_logic(0), s_real(0) {}
	const Block<T> &operator =(const Block<T>& r);
	void alloc(int n);
	void sort() { if (size()) qsort(0, size() - 1); }
	Block(const Block<T>& r) : data(null), s_logic(0), s_real(0)
	{
		(*this) = r;
	}
	int index(const T&x);
	void remove(int idx, int len = 1);
};

template <class T>
void Block<T>::remove(int idx, int len)
{
	int i, j;
	
	if (idx < 0 || len < 0 || idx + len > size())
		throw XBlock("bad remove() args");
	j = idx;
	for (i = idx + len; i < size(); i++, j++)
		data[j] = data[i];
	s_logic -= len;
}

template <class T>
int Block<T>::push(const T &r)
{
	int n = s_logic;
	
	alloc(s_logic + 1);
	
	data[n] = r;
	return n;
}

template <class T>
T Block<T>::pop()
{
	if (s_logic == 0)
		throw XBlock("pop() on empty");
	T r = data[--s_logic];
	if (s_logic == 0)
		free();
	return r;
}

template <class T>
const Block<T> &Block<T>::operator =(const Block<T> &r)
{
	free();
	s_logic = r.s_logic;
	s_real = r.s_real;
	data = new T[s_real];
	
	for (int i=0; i<s_logic; i++)
		data[i] = r.data[i];
	return *this;
}

template <class T>
void Block<T>::alloc(int k)
{
	if (!k) {
		free();
		return;
	}
	
	int old_logic = s_logic;
	s_logic = k;
	
	if (s_logic <= s_real)
		return;
		
	if (!s_real)
		s_real = BLOCK_MIN;
		
	while (s_real < s_logic)
		s_real *= 2;
	T *n = new T[s_real];
	
	for (int i=0; i<old_logic; i++)
		n[i] = data[i];
		
	delete [] data;
	data = n;
}

template <class T>
void Block<T>::qsort(int l, int r)
{
	int i, j;
	T x, x3;
	
	i = l; j = r;
	x = data[(l + r) / 2];
	do {
		while (data[i] < x)
			i++;
		while (x < data[j])
			j--;
		if (i <= j) {
			x3 = data[i];
			data[i] = data[j];
			data[j] = x3;
			i++;
			j--;
		}
	} while (i <= j);
	if (l < j)
		qsort(l, j);
	if (i < r)
		qsort(i, r);
}

template <class T>
int Block<T>::index(const T &x)
{
	for (int i = 0; i < size(); i++)
		if (x == data[i])
			return i;
	return -1;
}

#endif /* _V_BLOCK_H */

